// ttd.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#define VARDEF
#include "ttd.h"
#include "gfx.h"
#include "gui.h"
#include "station.h"
#include "vehicle.h"
#include "viewport.h"
#include "window.h"
#include "player.h"
#include "command.h"
#include "city.h"
#include "industry.h"
#include "news.h"
#include "engine.h"
#include "sound.h"
#include "economy.h"
#include "fileio.h"
#include "hal.h"

#include <stdarg.h>

//#define SPRITE_CACHE_SIZE 1024*1024*4 

void GameLoop();

void IncreaseSpriteLRU();
void InitializeGame();
void GenerateWorld(bool empty);
void CallLandscapeTick();
void IncreaseDate();
void RunOtherPlayersLoop();
void DoPaletteAnimations();
void MusicLoop();

extern void LoadMusicSettings();
extern void SaveMusicSettings();
extern void SetDifficultyLevel(int mode);
extern void DoStartupNewPlayer();

void redsq_debug(int tile);
bool read_language_pack(const char *filename);
bool LoadSavegame(const char *filename);

extern void HalGameLoop();

uint32 _pixels_redrawn;
bool _dbg_screen_rect;
bool disable_computer;

#if defined(WIN32)
void ShowWin32ErrorBox(const char *buf);
#endif

void CDECL error(const char *s, ...) {
	va_list va;
	char buf[512];
	va_start(va, s);
	vsprintf(buf, s, va);
	va_end(va);
	
#if defined(WIN32)
	ShowWin32ErrorBox(buf);
	if (_video_driver)
		_video_driver->stop();
#else
	if (_video_driver)
		_video_driver->stop();
	fprintf(stderr, "Error: %s\n", buf);
	assert(0);
#endif
	exit(1);
}

#if !defined(__GNUC__)
__int64 _declspec(naked) rdtsc()
{
	_asm {
		rdtsc
		ret
	}
}
#else
int rdtsc()
{
	return 0;
}
#endif


// NULL midi driver
static char *NullMidiStart(char **parm) { return NULL; }
static void NullMidiStop() {}
static void NullMidiPlaySong(const char *filename) {}
static void NullMidiStopSong() {}
static bool NullMidiIsSongPlaying() { return true; }
static void NullMidiSetVolume(byte vol) {}

static const HalMusicDriver _null_music_driver = {
	NullMidiStart,
	NullMidiStop,
	NullMidiPlaySong,
	NullMidiStopSong,
	NullMidiIsSongPlaying,
	NullMidiSetVolume,
};

// NULL video driver
static void *_null_video_mem;
static char *NullVideoStart(char **parm) {	_null_video_mem = malloc(_screen.width*_screen.height); return NULL; }
static void NullVideoStop() { free(_null_video_mem); }
static void NullVideoMakeDirty(int left, int top, int width, int height) {}
static int NullVideoMainLoop() {
	while(true) {
		GameLoop();
		_screen.dst_ptr = _null_video_mem;
		UpdateWindows();
	}	
}
	
static const HalVideoDriver _null_video_driver = {
	NullVideoStart,
	NullVideoStop,
	NullVideoMakeDirty,
	NullVideoMainLoop,
};

// NULL sound driver
static char *NullSoundStart(char **parm) { return NULL; }
static void NullSoundStop() {}
static const HalSoundDriver _null_sound_driver = {
	NullSoundStart,
	NullSoundStop,
};

enum {
	DF_PRIORITY_MASK = 0xf,
};

typedef struct {
	const char *name;
	const char *longname;
	const void *drv;
	uint flags;
} DriverDesc;


#if defined(WIN32)
extern HalVideoDriver *_win32_video_driver;
extern HalSoundDriver *_win32_sound_driver;
extern HalMusicDriver *_win32_music_driver;
#endif

#if defined(WITH_SDL)
extern HalSoundDriver *_sdl_sound_driver;
extern HalVideoDriver *_sdl_video_driver;
#endif

static const DriverDesc _video_driver_descs[] = {
	{"null", "Null Video Driver",				&_null_video_driver,	0},
#if defined(WITH_SDL)
	{"sdl", "SDL Video Driver",					&_sdl_video_driver,	1},
#endif
#if defined(WIN32)
	{"win32", "Win32 GDI Video Driver",	&_win32_video_driver,	2},
#endif
	{NULL}
};

static const DriverDesc _sound_driver_descs[] = {
	{"null", "Null Sound Driver",			&_null_sound_driver,	0},
#if defined(WITH_SDL)
	{"sdl", "SDL Sound Driver",				&_sdl_sound_driver,	1},
#endif
#if defined(WIN32)
	{"win32", "Win32 WaveOut Driver",	&_win32_sound_driver,	2},
#endif
	{NULL}
};

static const DriverDesc _music_driver_descs[] = {
	{"null", "Null Music Driver",		&_null_music_driver,	0},
#if defined(WIN32)
	{"win32", "Win32 MIDI Driver",	&_win32_music_driver,	1},
#endif
	{NULL}
};

typedef struct {
	const DriverDesc *descs;
	const char *name;
	void *var;
} DriverClass;

static DriverClass _driver_classes[] = {
	{_video_driver_descs, "video", &_video_driver},
	{_sound_driver_descs, "sound", &_sound_driver},
	{_music_driver_descs, "music", &_music_driver},
};

static const DriverDesc *GetDriverByName(const DriverDesc *dd, const char *name)
{
	do {
		if (!strcmp(dd->name, name))
			return dd;
	} while ((++dd)->name);
	return NULL;
}

static const DriverDesc *ChooseDefaultDriver(const DriverDesc *dd)
{
	const DriverDesc *best = NULL;
	int best_pri = -1;
	do {
		if ((int)(dd->flags&DF_PRIORITY_MASK) > best_pri) {
			best_pri = dd->flags&DF_PRIORITY_MASK;
			best = dd;
		}
	} while ((++dd)->name);
	return best;
}

void strlcpy(char *dst, const char *src, size_t len)
{
	assert(len >= 0);
	while (--len && *src)
		*dst++=*src++;
	*dst = 0;
}

void LoadDriver(int driver, const char *name)
{
	const DriverClass *dc = &_driver_classes[driver];
	const DriverDesc *dd;
	const void **var;
	const void *drv;
	const char *err;
	char *parm;
	char buffer[256];
	char *parms[32];

	parms[0] = NULL;

	if (name == NULL) {
		dd = ChooseDefaultDriver(dc->descs);
	} else {
		// Extract the driver name and put parameter list in parm
		strlcpy(buffer, name, sizeof(buffer));
		parm = strchr(buffer, ':');
		if (parm) {
			int np = 0;
			// Tokenize the parm.
			do {
				*parm++ = 0;
				if (np < lengthof(parms) - 1)
					parms[np++] = parm;
				while (*parm != 0 && *parm != ',')
					parm++;
			} while (*parm == ',');
			parms[np] = NULL;
		}
		dd = GetDriverByName(dc->descs, buffer);
		if (dd == NULL)
			error("No such %s driver: %s\n", dc->name, buffer);
	}
	var = dc->var;
	if (*var != NULL) ((HalCommonDriver*)*var)->stop();
	*var = NULL;
	drv = dd->drv;
	if ((err=((HalCommonDriver*)drv)->start(parms)) != NULL)
		error("Unable to load driver %s(%s). The error was: %s\n", dd->name, dd->longname, err);
	*var = drv;
}

void PrintDriverList()
{
	const DriverClass *dc = _driver_classes;
	const DriverDesc *dd;
	int i;
	for(i=0; i!=lengthof(_driver_classes); i++,dc++) { 
		printf("List of %s drivers:\n", dc->name);
		dd = dc->descs;
		do {
			printf("%10s: %s\n", dd->name, dd->longname);
		} while ((++dd)->name);
	}
}

char *GetDriverParam(char **parm, const char *name)
{
	char *p;
	int len = strlen(name);
	while (p = *parm++) {
		if (!strncmp(p,name,len)) {
			if (p[len] == '=') return p + len + 1;
			if (p[len] == 0)   return p + len;
		}
	}
	return NULL;
}

bool GetDriverParamBool(char **parm, const char *name)
{
	char *p = GetDriverParam(parm, name);
	return p != NULL;
}

int GetDriverParamInt(char **parm, const char *name, int def)
{
	char *p = GetDriverParam(parm, name);
	return p != NULL ? atoi(p) : def;
}

typedef struct {
	char *opt;
	int numleft;
	char **argv;
	const char *options;
	char *cont;
} MyGetOptData;

static void MyGetOptInit(MyGetOptData *md, int argc, char **argv, const char *options)
{
	md->cont = NULL;
	md->numleft = argc;
	md->argv = argv;
	md->options = options;
}

static int MyGetOpt(MyGetOptData *md)
{
	char *s,*r,*t;

	if ((s=md->cont) != NULL)
		goto md_continue_here;

	while(true) {
		if (--md->numleft < 0)
			return -1;

		s = *md->argv++;
		if (*s == '-') {
md_continue_here:;
			s++;
			if (*s != 0) {
				// Found argument, try to locate it in options.
				if (*s == ':' || (r = strchr(md->options, *s)) == NULL) {
					// ERROR!
					return -2;
				}
				if (r[1] == ':') {
					// Item wants an argument. Check if the argument follows, or if it comes as a separate arg.
					if (!*(t = s + 1)) {
						// It comes as a separate arg. Check if out of args?
						if (--md->numleft < 0 || *(t = *md->argv) == '-') {
							// Check if item is optional?
							if (r[2] != ':')
								return -2;
							t = NULL;
						} else {
							md->argv++;
						}
					}
					md->opt = t;
					md->cont = NULL;
					return *s;
				}
				md->opt = NULL;
				md->cont = s;
				return *s;
			}
		} else {
			// This is currently not supported.
			return -2;
		}
	}
}

static uint16 _screen_sizes[] = {
	640, 800, 1024,
	480, 600, 768
};

//#define WIN32_EXCEPTION_TRACKER_DEBUG
//#define WIN32_EXCEPTION_TRACKER

int CDECL main(int argc, char* argv[])
{
	MyGetOptData mgo;
	int i;
	char *video_driver = NULL;
	char *sound_driver = NULL;
	char *music_driver = NULL;
	int res = 0;
	int network = 0;
	char *network_conn;

#if defined(WIN32_EXCEPTION_TRACKER)
	{	
		extern void Win32InitializeExceptions();
		Win32InitializeExceptions();
	}
#endif

#if defined(WIN32_EXCEPTION_TRACKER_DEBUG)
	_try {
		uint32 _stdcall ExceptionHandler(void *ep);
#endif

	_in_scenario_editor = 0;

	MyGetOptInit(&mgo, argc-1, argv+1, "m:s:v:h81n::l:e");
	while ((i = MyGetOpt(&mgo)) != -1) {
		switch(i) {
		case 'm': music_driver = mgo.opt; break;
		case 's': sound_driver = mgo.opt; break;
		case 'v': video_driver = mgo.opt; break;
		case '8':	res = 1; break;		
		case '1': res = 2; break;
		case 'n': network=1; if (mgo.opt) {network_conn = mgo.opt; network++;} break;
		case 'l': {
				int k = atoi(mgo.opt);
				if (IS_INT_INSIDE(k, 1, 5))
					_opt.landscape = k - 1;
		} break;
		case 'e': _in_scenario_editor = 2 ; break;
		case -2:
			printf("Invalid command line.\n");
		case 'h':
			printf(
				"Command line options:\n"
				"  -v drv = Set video driver (see below)\n"
				"  -s drv = Set sound driver (see below)\n"
				"  -m drv = Set music driver (see below)\n"
				"  -8     = Use 800x600 resolution\n"
				"  -1     = Use 1024x768 resolution\n"
				"  -h     = Display this help text\n"
				"  -l 1-4 = Use this landscape\n"
				"  -e     = Start Editor\n"
			);
			PrintDriverList();
			return 0;
		}
	}

	_screen.pitch = _screen.width = _screen_sizes[res];
	_screen.height = _screen_sizes[res+3];

	// Load language-specific strings.
	if (!read_language_pack("english.lng")) {
		error("Unable to read english.lng");
	}

	// Sample catalogue
	puts("Loading sound effects...");
	MxInitialize(11025, "sample.cat"); 

	// This must be done early, since functions use the InvalidateWindow* calls
	InitWindowSystem();

	GfxLoadSprites();
	LoadStringWidthTable();
	
	puts("Loading drivers...");
	LoadDriver(VIDEO_DRIVER, video_driver);
	LoadDriver(SOUND_DRIVER, sound_driver);
	LoadMusicSettings();
	LoadDriver(MUSIC_DRIVER, music_driver);
	MusicLoop();

	// Load string widths
	

	_num_human_players = 1;
	_display_opt = 0xff & ~DO_FULL_ANIMATION;
	_news_display_opt = -1;
	
	// Default difficulty level
	_opt_mod_ptr = &_new_opt;
	SetDifficultyLevel(0);

	_switch_mode = 4;

#define PORTNUM 1234
	if (network) {
		DoStartupNewPlayer();
		_players[1].player_money = 1000000;
		_num_human_players = 2;
		_human_player_2 = 1;
		
		NetworkInitialize();
		if (network==1) {
			printf("Listening on port %d\n", PORTNUM);
			NetworkListen(PORTNUM);
		} else {
			printf("Connecting to %s %d\n", network_conn, PORTNUM);
			NetworkConnect(network_conn,PORTNUM);
		}
	}

	while (_video_driver->main_loop() == ML_SWITCHDRIVER) {}

	_video_driver->stop();
	_music_driver->stop();
	_sound_driver->stop();

	SaveMusicSettings();

#if defined(WIN32_EXCEPTION_TRACKER_DEBUG)
	} _except (ExceptionHandler(_exception_info())) {}
#endif

	return 0;
}

static void SwitchMode(int new_mode)
{
	_in_state_game_loop = true;
	
	switch(new_mode) {
	case 2: // Switch to scenario editor
		_in_scenario_editor = 2;

		// Copy in game options
		_opt_mod_ptr = &_opt;
		memcpy(&_opt, &_new_opt, sizeof(_opt));
		
		GfxLoadSprites();

		// Re-init the windowing system
		InitWindowSystem();

		// Create toolbars
		SetupColorsAndInitialWindow();

		// Startup the game system
		GenerateWorld(true);

		_human_player_1 = 0x10;
		MarkWholeScreenDirty();
		break;

	case 1: // Make new world
		_in_scenario_editor = 1;

		// Copy in game options
		_opt_mod_ptr = &_opt;
		memcpy(&_opt, &_new_opt, sizeof(_opt));

		GfxLoadSprites();

		// Reinitialize windows
		InitWindowSystem();
		LoadStringWidthTable();

		SetupColorsAndInitialWindow();
		
		// Randomize world
		GenerateWorld(false);
		
		// Create a single player
		DoStartupNewPlayer();
		_human_player_1 = 0;
		MarkWholeScreenDirty();
		break;

	case 3: // Load game
		_in_scenario_editor = 1;

		_human_player_1 = 0;
		SaveOrLoad(_file_to_saveload, false);

		break;

	case 4: // Switch to game menu
		_in_scenario_editor = 0;

		_opt_mod_ptr = &_new_opt;
		GfxLoadSprites();
		LoadStringWidthTable();
		// Setup main window
		InitWindowSystem();
		SetupColorsAndInitialWindow();

		// Generate a world.
		GenerateWorld(false);
		_human_player_1 = 0;
		MarkWholeScreenDirty();
		break;

	case 5: // Save game
		if (!SaveOrLoad(_file_to_saveload, true))
			ShowErrorMessage(INVALID_STRING_ID, STR_4007_GAME_SAVE_FAILED, 0, 0);
		else
			DeleteWindowById(WC_SAVELOAD, 0);
		break;
	}

	_in_state_game_loop = false;
}


// State controlling game loop.
// The state must not be changed from anywhere
// but here.
// That check is enforced in DoCommand.
void StateGameLoop()
{
	_in_state_game_loop = true;
	_frame_counter++;
	
	if (_in_scenario_editor == 2) {
		RunTileLoop();
		CallVehicleTicks();
		CallLandscapeTick();
		CallWindowTickEvent();
		NewsLoop();
	} else {
		AnimateAnimatedTiles();
		IncreaseDate();
		RunTileLoop();
		CallVehicleTicks();
		CallLandscapeTick();
		if (!disable_computer)
			RunOtherPlayersLoop();

		CallWindowTickEvent();
		NewsLoop();
	}
	_in_state_game_loop = false;
}

//
void GameLoop()
{
	
	if (_switch_mode) {
		int m = _switch_mode;
		_switch_mode = 0;
		SwitchMode(m);
	}
	IncreaseSpriteLRU();
	InteractiveRandom();

	if (_scroller_click_timeout > 3)
		_scroller_click_timeout -= 3;
	else
		_scroller_click_timeout = 0;

	_caret_timer += 3;
	_timer_counter+=8;
	CursorTick();

	NetworkPoll();
	RecorderPoll();

	if (_game_paused == 0)
		StateGameLoop();
	
	MouseLoop();
	MoveAllTextEffects();

	if (_display_opt & DO_FULL_ANIMATION)
		DoPaletteAnimations();

	if (_in_scenario_editor != 0)
		MusicLoop();
}

void BeforeSaveGame()
{
	Window *w = FindWindowById(WC_MAIN_WINDOW, 0);

	_saved_scrollpos_x = WP(w,vp_d).scrollpos_x;
	_saved_scrollpos_y = WP(w,vp_d).scrollpos_y;
	_saved_scrollpos_zoom = w->viewport->zoom;
}

void AfterLoadGame()
{
	Window *w;
	int diff;

	// Load the sprites
	GfxLoadSprites();

	// Update current year
	{
		YearMonthDay ymd;
		ConvertDayToYMD(&ymd, _date);
		_cur_month = ymd.month;
		_cur_year = ymd.year;
	}

	// Update all vehicles
	AllVehiclePositionChanged();

	// Initialize windows
	_opt_mod_ptr = &_opt;
	InitWindowSystem();
	SetupColorsAndInitialWindow();

	w = FindWindowById(WC_MAIN_WINDOW, 0);

	WP(w,vp_d).scrollpos_x = _saved_scrollpos_x;
	WP(w,vp_d).scrollpos_y = _saved_scrollpos_y;
	
	diff = _saved_scrollpos_zoom - w->viewport->zoom;
	w->viewport->zoom = _saved_scrollpos_zoom;
	if (diff < 0) {
		w->viewport->virtual_width >>= (-diff);
		w->viewport->virtual_height >>= (-diff);
	} else if (diff > 0) {
		w->viewport->virtual_width <<= diff;
		w->viewport->virtual_height <<= diff;
	}
	
	UpdateZoomButtonState();

	MarkWholeScreenDirty();
}


void DebugProc(int i)
{
	if (i == 0) {
		*(byte*)0 = 0;
	} else {
		_players[0].player_money += 100000;
		MarkWholeScreenDirty();
	}

}

void PauseGame()
{
	_game_paused++;
	InvalidateWindow(WC_STATUS_BAR, 0);
}

void UnpauseGame()
{
	_game_paused--;
	InvalidateWindow(WC_STATUS_BAR, 0);
}